"""
"""

# This file is part of genetest
#
# This work is licensed under the Creative Commons Attribution-NonCommercial
# 4.0 International License. To view a copy of this license, visit
# http://creativecommons.org/licenses/by-nc/4.0/ or send a letter to Creative
# Commons, PO Box 1866, Mountain View, CA 94042, USA.


import os
import sys
import shlex
import logging
import argparse
from os import path

import pandas as pd

from .. import modelspec as spec
from .. import __version__, _LOG_FORMAT
from ..modelspec.predicates import NameFilter
from ..modelspec import modelspec_from_formula
from ..analysis import execute, execute_formula
from ..configuration import AnalysisConfiguration
from ..modelspec import result as analysis_results
from ..phenotypes.dataframe import DataFrameContainer
from ..subscribers import GWASWriter, ResultsMemory, RowWriter


_TESTING_MODE = False


# Configuring the log
logger = logging.getLogger("genetest")


def main():
    """The main function."""
    # Creating the option parser
    desc = ("Performs statistical analysis on genotypic data "
            "(version {}).".format(__version__))
    parser = argparse.ArgumentParser(description=desc)

    # The logging file handler
    logging_fh = None

    # Parsing the options
    args = parse_args(parser)

    try:
        # Checking the options
        check_args(args)

        if not _TESTING_MODE:
            sh = logging.StreamHandler()
            sh.setFormatter(_LOG_FORMAT)
            logger.addHandler(sh)
            logger.setLevel("INFO")

            # Adding the file handler to the logger
            logging_fh = logging.FileHandler(args.output + ".log", mode="w")
            logging_fh.setFormatter(_LOG_FORMAT)
            logger.addHandler(logging_fh)

        # Logging
        logger.info("This is {script_name} version {version}".format(
            script_name=os.path.basename(sys.argv[0]),
            version=__version__,
        ))
        logger.info("Arguments: {}".format(
            " ".join(shlex.quote(part) for part in sys.argv[1:]),
        ))

        # Getting the analysis configuration
        logger.info("Parsing configuration file")
        conf = AnalysisConfiguration(args.configuration)

        # Getting the phenotypes
        logger.info("Creating phenotypes container")
        phenotypes = conf.get_phenotypes()
        logger.info("  - {:,d} samples, {:,d} variables".format(
            phenotypes.get_nb_samples(),
            phenotypes.get_nb_variables(),
        ))
        if args.keep:
            phenotypes.keep_samples(set(args.keep.read().splitlines()))
            args.keep.close()
            logger.info(
                "  - {:,d} samples kept".format(phenotypes.get_nb_samples()),
            )

        # Getting the genotypes
        logger.info("Creating genotypes container")
        genotypes = conf.get_genotypes()
        logger.info("  - {:,d} samples, {:,d} markers".format(
            genotypes.get_number_samples(),
            genotypes.get_number_variants(),
        ))

        # Creating a list of variant predicates
        variant_predicates = []

        # The markers to extract
        if args.extract:
            to_extract = set(args.extract.read().splitlines())
            args.extract.close()

            logger.info("{:,d} variants will be extracted".format(
                len(to_extract),
            ))
            variant_predicates.append(NameFilter(extract=to_extract))

        # Checking the test information
        test = conf.get_model_test()
        formula = conf.get_model_formula()
        test_kwargs = conf.get_model_args()
        logger.info("Analysis: {}".format(test))
        logger.info("  - {}".format(formula))
        for k, v in test_kwargs.items():
            logger.info("  - {}: {}".format(k, v))

        # Checking for special test kwargs
        mixedlm_optimization = test_kwargs.get("optimize", True)
        mixedlm_p_threshold = test_kwargs.get("p_threshold", 1e-4)
        for to_del in ("optimize", "p_threshold"):
            if to_del in test_kwargs:
                del test_kwargs[to_del]

        if test == "mixedlm" and mixedlm_optimization:
            # Performing an "optimized" mixed linear model
            performed_optimized_mixedlm(
                args=args, test=test, phenotypes=phenotypes,
                genotypes=genotypes, formula=formula, test_kwargs=test_kwargs,
                variant_predicates=variant_predicates, p_t=mixedlm_p_threshold,
            )

        else:
            # Performing a "normal" analysis
            perform_normal_analysis(
                args=args, test=test, phenotypes=phenotypes,
                genotypes=genotypes, formula=formula, test_kwargs=test_kwargs,
                variant_predicates=variant_predicates,
            )

    # Catching the Ctrl^C
    except KeyboardInterrupt:
        logger.info("Cancelled by user")
        sys.exit(0)

    # Catching the CliError
    except CliError as e:
        logger.critical(e.message)
        parser.error(e.message)

    finally:
        # Closing the logging file
        if logging_fh:
            logging_fh.close()

        # Closing the "keep" file
        if args.keep and not args.keep.closed:
            args.keep.close()

        # Closing the "extract" file
        if args.extract and not args.extract.closed:
            args.extract.close()


def performed_optimized_mixedlm(args, test, phenotypes, genotypes, formula,
                                test_kwargs, variant_predicates, p_t):
    """Performs an "optimized" mixed linear model."""
    logger.info("Optimizing MixedLM analysis")

    # Getting the model specification and the subgroups (if any)
    modelspec, subgroups = modelspec_from_formula(formula, test, test_kwargs)

    # TODO: Find out if this is possible
    if subgroups is not None:
        raise CliError("Subgroups are not available for MixedLM optimization")

    # Removing the SNPs from the predictors
    if "SNPs" in modelspec.predictors:
        del modelspec.predictors[modelspec.predictors.index("SNPs")]

    # Executing the normal MixedLM analysis and executing the analysis
    logger.info("Computing the random effects")
    memory_subscriber = ResultsMemory()
    execute(
        phenotypes=phenotypes, genotypes=genotypes, modelspec=modelspec,
        subscribers=[memory_subscriber], output_prefix=args.output,
        subgroups=subgroups, maf_t=args.maf, cpus=args.nb_cpus,
    )

    # Getting the RE values
    random_effects = memory_subscriber.results.pop()["MODEL"]["random_effects"]

    # Getting the column for the groupings
    group_col = modelspec.get_translations()[modelspec.outcome["groups"].id]

    # Checking we don't have the random effects column in the original pheno
    assert "_random_effects" not in phenotypes.get_phenotypes().columns

    # We want to keep only the unique samples of the old phenotypes
    old_phenotypes = phenotypes.get_phenotypes()
    duplicated = old_phenotypes.index.duplicated(keep="first")

    # Merging the random effects to the phenotypes
    new_phenotypes = pd.merge(
        left=old_phenotypes.loc[~duplicated, :].copy(),
        right=pd.DataFrame(random_effects, columns=["_random_effects"]),
        left_on=group_col,
        right_index=True,
    )

    # Resetting the model spec
    spec._reset()

    # Creating the new phenotypes container
    new_phenotypes = DataFrameContainer(dataframe=new_phenotypes)
    assert not new_phenotypes.is_repeated()

    # Creating a new model spec
    optimized_modelspec = spec.ModelSpec(
        outcome=spec.phenotypes._random_effects,
        predictors=[spec.SNPs],
        test="linear",
    )

    # Creating a row subscriber
    approximation_fn = args.output + ".mixedlm_approximation.txt"
    subscribers = [RowWriter(
        filename=approximation_fn,
        columns=[
            ("snp", analysis_results["SNPs"]["name"]),
            ("chr", analysis_results["SNPs"]["chrom"]),
            ("pos", analysis_results["SNPs"]["pos"]),
            ("major", analysis_results["SNPs"]["major"]),
            ("minor", analysis_results["SNPs"]["minor"]),
            ("maf", analysis_results["SNPs"]["maf"]),
            ("n", analysis_results["MODEL"]["nobs"]),
            ("p", analysis_results["SNPs"]["p_value"]),
            ("ll", analysis_results["MODEL"]["log_likelihood"]),
            ("adj_r2", analysis_results["MODEL"]["r_squared_adj"]),
            ("approximation", "Yes")
        ],
        header=True,
        append=False,
    )]

    # Executing the optimized analysis
    logger.info("Executing the optimized MixedLM")
    execute(
        phenotypes=new_phenotypes, genotypes=genotypes,
        modelspec=optimized_modelspec, subscribers=subscribers,
        output_prefix=args.output + ".mixedlm_approximation",
        maf_t=args.maf, cpus=args.nb_cpus,
        variant_predicates=variant_predicates,
    )

    # Making sure the output file exists
    if not path.isfile(approximation_fn):
        raise CliError("{}: no such file".format(approximation_fn))

    # Reading the approximation
    approximation = pd.read_csv(approximation_fn, sep="\t")

    # Keeping only small p values (according to threshold)
    markers = set(approximation.loc[approximation.p < p_t, "snp"].values)

    if len(markers) == 0:
        logger.info("No marker had a p value < than {}".format(p_t))
        return

    # Adding the SNPs into the original ModelSpec
    modelspec.predictors.append("SNPs")

    # Creating the new variant predicates
    variant_predicates = [NameFilter(extract=markers)]

    # Creating the GWAS subscriber
    subscribers = [GWASWriter(filename=args.output + ".txt", test="mixedlm")]

    # Resetting the model spec
    spec._reset()

    # Executing the real MixdLM
    logger.info("Executing MixedLM on {:,d} markers".format(len(markers)))
    execute_formula(
        phenotypes=phenotypes, genotypes=genotypes, formula=formula,
        test="mixedlm", test_kwargs=test_kwargs, subscribers=subscribers,
        variant_predicates=variant_predicates, output_prefix=args.output,
        maf_t=args.maf, cpus=args.nb_cpus,
    )


def perform_normal_analysis(args, test, phenotypes, genotypes, formula,
                            test_kwargs, variant_predicates):
    """Performs a "normal" analysis."""
    # Creating a GWAS subscriber
    subscribers = [GWASWriter(filename=args.output + ".txt", test=test)]

    # Starting the analysis
    execute_formula(
        phenotypes=phenotypes, genotypes=genotypes, formula=formula, test=test,
        test_kwargs=test_kwargs, subscribers=subscribers,
        variant_predicates=variant_predicates, output_prefix=args.output,
        maf_t=args.maf, cpus=args.nb_cpus,
    )


def check_args(args):
    """Checks the arguments and options."""
    # Checking that configuration file exists
    if not os.path.isfile(args.configuration):
        raise CliError("{}: no such file.".format(args.configuration))

    # Checking the number of CPUs
    if args.nb_cpus < 1:
        raise CliError("{}: invalid number of CPUs".format(args.nb_cpus))

    # Checking the MAF
    if args.maf < 0 or args.maf > 0.5:
        raise CliError("{}: invalid MAF".format(args.maf))


def parse_args(parser):
    """Parses the command line options and arguments."""
    parser.add_argument("-v", "--version", action="version",
                        version="%(prog)s {}".format(__version__))
    parser.add_argument("--test", action=TestAction, nargs=0,
                        help="Execute the test suite and exit.")
    parser.add_argument(
        "--nb-cpus", type=int, metavar="NB", default=1,
        help="The number of processes to use for the analysis. [%(default)d]",
    )

    # The input options
    group = parser.add_argument_group("Input Options")
    group.add_argument(
        "--configuration", type=str, metavar="INI", required=True,
        help="The configuration file that describe the phenotypes, genotypes, "
             "and model.",
    )

    # The output options
    group = parser.add_argument_group("Output Options")
    group.add_argument(
        "--output", type=str, metavar="FILE", default="genetest_results",
        help="The output file prefix that will contain the results and other "
             "information. [%(default)s]",
    )

    # Some other options
    group = parser.add_argument_group("Other Options")
    group.add_argument(
        "--extract", type=argparse.FileType("r"), metavar="FILE",
        help="A file containing a list of marker to extract prior to the "
             "statistical analysis (one marker per line).",
    )
    group.add_argument(
        "--keep", type=argparse.FileType("r"), metavar="FILE",
        help="A file containing a list of samples to keep prior to the "
             "statistical analysis (one sample per line).",
    )
    group.add_argument(
        "--maf", type=float, default=0.01, metavar="MAF",
        help="The MAF threshold to include a marker in the analysis. "
             "[%(default).2f]",
    )

    return parser.parse_args()


class TestAction(argparse.Action):  # pragma: no cover
    def __call__(self, parser, namespace, values, option_string=None):
        import unittest
        from ..tests import test_suite
        unittest.TextTestRunner(verbosity=2).run(test_suite)
        parser.exit()


class CliError(Exception):
    """An Exception raised in case of a problem."""
    def __init__(self, msg):
        """Construction of the CliError class."""
        self.message = str(msg)

    def __str__(self):
        return self.message

    def __repr__(self):
        return "CliError: " + self.message





# This file is part of genetest.
#
# This work is licensed under the Creative Commons Attribution-NonCommercial
# 4.0 International License. To view a copy of this license, visit
# http://creativecommons.org/licenses/by-nc/4.0/ or send a letter to Creative
# Commons, PO Box 1866, Mountain View, CA 94042, USA.


import os
import unittest
from tempfile import TemporaryDirectory

from ..configuration import AnalysisConfiguration


__copyright__ = "Copyright 2016, Beaulieu-Saucier Pharmacogenomics Centre"
__license__ = "Attribution-NonCommercial 4.0 International (CC BY-NC 4.0)"


class TestAnalysisConfiguration(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.tmpdir = TemporaryDirectory(prefix="genetest_")

    @classmethod
    def tearDownClass(cls):
        cls.tmpdir.cleanup()

    """Tests the 'AnalysisConfiguration' module."""
    def test_missing_section(self):
        """Tests when there is a missing section."""
        conf = os.path.join(self.tmpdir.name, "conf.yaml")
        with open(conf, "w") as f:
            print(
                "phenotypes:\n"
                "    format: text\n"
                "    filename: dummy_file\n"
                "genotypes:\n"
                "    format: plink\n"
                "    prefix: dummy_prefix",
                file=f,
            )

        with self.assertRaises(ValueError) as cm:
            AnalysisConfiguration(conf)

        self.assertTrue(str(cm.exception).startswith("Missing section"))

    def test_extra_section(self):
        """Tests when there is an extra section."""
        conf = os.path.join(self.tmpdir.name, "conf.yaml")
        with open(conf, "w") as f:
            print(
                "phenotypes:\n"
                "    format: text\n"
                "    filename: dummy_file\n"
                "genotypes:\n"
                "    format: plink\n"
                "    prefix: dummy_prefix\n"
                "model:\n"
                "    test: linear\n"
                "    formula: pheno~age\n"
                "dummy:\n"
                "    format:none",
                file=f,
            )

        with self.assertRaises(ValueError) as cm:
            AnalysisConfiguration(conf)

        self.assertTrue(str(cm.exception).startswith("Invalid section"))

    def test_missing_phenotype_format(self):
        """Tests when the phenotype format is missing."""
        conf = os.path.join(self.tmpdir.name, "conf.yaml")
        with open(conf, "w") as f:
            print(
                "phenotypes:\n"
                "    filename: dummy_file\n"
                "genotypes:\n"
                "    format: plink\n"
                "    prefix: dummy_prefix\n"
                "model:\n"
                "    test: linear\n"
                "    formula: pheno~age",
                file=f,
            )

        with self.assertRaises(ValueError) as cm:
            AnalysisConfiguration(conf)

        self.assertTrue(
            "the 'phenotypes' section should contain the "
            "'format'" in str(cm.exception)
        )

    def test_invalid_phenotype_format(self):
        """Tests when an invalid format for phenotypes is asked."""
        conf = os.path.join(self.tmpdir.name, "conf.yaml")
        with open(conf, "w") as f:
            print(
                "phenotypes:\n"
                "    format: nope\n"
                "    filename: dummy_file\n"
                "genotypes:\n"
                "    format: plink\n"
                "    prefix: dummy_prefix\n"
                "model:\n"
                "    test: linear\n"
                "    formula: pheno~age",
                file=f,
            )

        with self.assertRaises(ValueError) as cm:
            AnalysisConfiguration(conf)

        self.assertTrue(str(cm.exception).startswith(
            "Invalid 'format' (nope) for the 'phenotypes' section"
        ))

    def test_missing_genotype_format(self):
        """Tests when the genotype format is missing."""
        conf = os.path.join(self.tmpdir.name, "conf.yaml")
        with open(conf, "w") as f:
            print(
                "phenotypes:\n"
                "    format: text\n"
                "    filename: dummy_file\n"
                "genotypes:\n"
                "    prefix: dummy_prefix\n"
                "model:\n"
                "    test: linear\n"
                "    formula: pheno~age",
                file=f,
            )

        with self.assertRaises(ValueError) as cm:
            AnalysisConfiguration(conf)

        self.assertTrue(
            "the 'genotypes' section should contain the "
            "'format'" in str(cm.exception)
        )

    def test_invalid_genotype_format(self):
        """Tests when an invalid format for genotypes is asked."""
        conf = os.path.join(self.tmpdir.name, "conf.yaml")
        with open(conf, "w") as f:
            print(
                "phenotypes:\n"
                "    format: text\n"
                "    filename: dummy_file\n"
                "genotypes:\n"
                "    format: nope\n"
                "    prefix: dummy_prefix\n"
                "model:\n"
                "    test: linear\n"
                "    formula: pheno~age",
                file=f,
            )

        with self.assertRaises(ValueError) as cm:
            AnalysisConfiguration(conf)

        self.assertTrue(str(cm.exception).startswith(
            "Invalid 'format' (nope) for the 'genotypes' section"
        ))

    def test_missing_test(self):
        """Tests when the statistics test is missing."""
        conf = os.path.join(self.tmpdir.name, "conf.yaml")
        with open(conf, "w") as f:
            print(
                "phenotypes:\n"
                "    format: text\n"
                "    filename: dummy_file\n"
                "genotypes:\n"
                "    format: plink\n"
                "    prefix: dummy_prefix\n"
                "model:\n"
                "    formula: pheno~age",
                file=f,
            )

        with self.assertRaises(ValueError) as cm:
            AnalysisConfiguration(conf)

        self.assertTrue(
            "the 'model' section should contain the "
            "'test'" in str(cm.exception)
        )

    def test_invalid_test(self):
        """Tests when an invalid model for statistics is asked."""
        conf = os.path.join(self.tmpdir.name, "conf.yaml")
        with open(conf, "w") as f:
            print(
                "phenotypes:\n"
                "    format: text\n"
                "    filename: dummy_file\n"
                "genotypes:\n"
                "    format: plink\n"
                "    prefix: dummy_prefix\n"
                "model:\n"
                "    test: nope\n"
                "    formula: pheno~age",
                file=f,
            )

        with self.assertRaises(ValueError) as cm:
            AnalysisConfiguration(conf)

        self.assertTrue(str(cm.exception).startswith(
            "Invalid 'test' for the 'model' section"
        ))

    def test_normal_functionality(self):
        """Tests the normal functionality."""
        conf = os.path.join(self.tmpdir.name, "conf.yaml")
        with open(conf, "w") as f:
            print(
                "phenotypes:\n"
                "    format: text\n"
                "    filename: dummy_file\n"
                "    options:\n"
                "        field_separator: \\t\n"
                "        repeated_measurements: No\n"
                "        sample_column: sample_id\n"
                "        missing_values: ['-9', '-999']\n"
                "genotypes:\n"
                "    format: impute2\n"
                "    filename: fn\n"
                "    sample_filename: sample_fn\n"
                "    options:\n"
                "        probability_threshold: 0.95\n"
                "model:\n"
                "    test: linear\n"
                "    formula: y~x+z\n"
                "    options:\n"
                "        condition_value_t: 3000",
                file=f,
            )

        # Getting the configuration
        conf = AnalysisConfiguration(conf)

        # Checking the genotypes
        observed_geno_args = conf.get_genotypes_args()
        self.assertEqual(3, len(observed_geno_args))
        self.assertEqual("fn", observed_geno_args["filename"])
        self.assertEqual("sample_fn", observed_geno_args["sample_filename"])
        self.assertEqual(0.95, observed_geno_args["probability_threshold"])
        self.assertEqual("impute2", conf.get_genotypes_format())

        # Checking the phenotypes
        observed_pheno_args = conf.get_phenotypes_args()
        self.assertEqual(5, len(observed_pheno_args))
        self.assertEqual("dummy_file", observed_pheno_args["filename"])
        self.assertEqual("sample_id", observed_pheno_args["sample_column"])
        self.assertEqual("\t", observed_pheno_args["field_separator"])
        self.assertEqual(["-9", "-999"], observed_pheno_args["missing_values"])
        self.assertFalse(observed_pheno_args["repeated_measurements"])
        self.assertEqual("text", conf.get_phenotypes_format())

        # Checking the statistics
        observed_stats_args = conf.get_model_args()
        self.assertEqual(1, len(observed_stats_args))
        self.assertEqual(3000, observed_stats_args["condition_value_t"])
        self.assertEqual("y~x+z", conf.get_model_formula())
        self.assertEqual("linear", conf.get_model_test())



#!/usr/bin/env python

# How to build source distribution
#   - python setup.py sdist --format bztar
#   - python setup.py sdist --format gztar
#   - python setup.py sdist --format zip
#   - python setup.py bdist_wheel


import os
import sys
import subprocess

from setuptools import setup, find_packages


MAJOR = 0
MINOR = 2
MICRO = 1
VERSION = "{0}.{1}.{2}".format(MAJOR, MINOR, MICRO)


def check_python_version():
    """Checks the python version, exits if < 3.3."""
    python_major, python_minor = sys.version_info[:2]

    if python_major != 3 or python_minor < 3:
        sys.stderr.write("genetest requires python 3 "
                         "(version 3.3 or higher)\n")
        sys.exit(1)


def write_version_file(fn=None):
    if fn is None:
        fn = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            os.path.join("genetest", "version.py"),
        )

    content = ("\n# THIS FILE WAS GENERATED AUTOMATICALLY\n"
               'genetest_version = "{version}"\n')

    a = open(fn, "w")
    try:
        a.write(content.format(version=VERSION))
    finally:
        a.close()


def build_grako_parser():
    base = os.path.abspath(
        os.path.join(
            os.path.dirname(__file__), "genetest", "modelspec"
        )
    )
    target = os.path.join(base, "parser.py")

    grammar = os.path.join(base, "modelspec.ebnf")

    args = [
        "python", "-m", "grako", grammar, "-o", target
    ]
    print("Building modelspec grammar parser.")
    print(" ".join(args))
    subprocess.check_call(args)


def setup_package():
    # Checking the python version prior to installation
    check_python_version()

    # Saving the version into a file
    write_version_file()

    # Build the grecko parser.
    build_grako_parser()

    setup(
        name="genetest",
        version=VERSION,
        description="A package to process and analyze genotypic and "
                    "phenotypic data.",
        long_description="This package provides tools to process and analyse "
                         "genotypic and phenotypic data using different "
                         "statistical models. The 'genotypes' and "
                         "'phenotypes' sub-modules provide interfaces to "
                         "process genotypes and phenotypes in different "
                         "format. The 'statistics' sub-module will interface "
                         "with them to analyze the data using different "
                         "statistical models.",
        url="https://github.com/pgxcentre/genetest",
        license="CC BY-NC 4.0",
        test_suite="genetest.tests.test_suite",
        zip_safe=False,
        entry_points={
            "console_scripts": [
                "genetest=genetest.scripts.cli:main",
            ],
        },
        install_requires=["numpy >= 1.12.0", "pandas >= 0.19.0",
                          "setuptools >= 26.1.0", "statsmodels >= 0.8.0",
                          "grako >= 3.10.0", "scipy >= 0.19.0",
                          "geneparse >= 0.1.0", "pyyaml >= 3.12"],
        packages=find_packages(),
        package_data={"genetest.tests": ["data/genotypes/*",
                                         "data/statistics/*"]},
        classifiers=["Development Status :: 4 - Beta",
                     "Intended Audience :: Science/Research",
                     "License :: Free for non-commercial use",
                     "Operating System :: Unix",
                     "Operating System :: POSIX :: Linux",
                     "Operating System :: MacOS :: MacOS X",
                     "Operating System :: Microsoft",
                     "Programming Language :: Python",
                     "Programming Language :: Python :: 3.3",
                     "Programming Language :: Python :: 3.4",
                     "Programming Language :: Python :: 3.5",
                     "Topic :: Scientific/Engineering :: Bio-Informatics"],
        keywords="bioinformatics genetics statistics",
    )


if __name__ == "__main__":
    setup_package()



"""
Semantics for the grako parser.
"""

import logging
import warnings

from .core import (SNPs, interaction, phenotypes, genotypes, factor, pow,
                   log10, ln)


logger = logging.getLogger(__name__)


try:
    from .parser import ModelSpecParser
    PARSER_AVAIL = True
except ImportError as e:
    logger.warning("No parser available: " + str(e))
    PARSER_AVAIL = False


class ModelSpecSemantics(object):
    def __init__(self):
        self.entities = {}

    def SNPs(self, ast):
        return SNPs

    def labelled_outcome_group(self, ast):
        return {tree["key"]: tree["name"] for tree in ast["tags"]}

    def interaction(self, ast):
        return interaction(*ast["interaction"], name=ast["as_"])

    def phenotype(self, ast):
        return phenotypes[ast["name"]]

    def integer(self, ast):
        return int(ast["int"])

    def genotype(self, ast):
        return genotypes[ast["variant"]]

    def factor(self, ast):
        return factor(ast["phen"], name=ast["as_"])

    def pow(self, ast):
        return pow(ast["phen"], ast["power"], name=ast["as_"])

    def ln(self, ast):
        return ln(ast["phen"])

    def log10(self, ast):
        return log10(ast["phen"])

    def _default(self, ast):
        return ast


def parse_modelspec(s):
    """Use the modelspec grammar to parse kwargs for modelspec."""
    warnings.warn("use 'parse_formula' instead", DeprecationWarning)
    return parse_formula(s)


def parse_formula(f):
    """Use the modelspec grammar to parse a formula for the modelspec."""
    if not PARSER_AVAIL:
        return

    parser = ModelSpecParser(parseinfo=False)
    return parser.parse(f, rule_name="model", semantics=ModelSpecSemantics())





# This file is part of genetest.
#
# This work is licensed under the Creative Commons Attribution-NonCommercial
# 4.0 International License. To view a copy of this license, visit
# http://creativecommons.org/licenses/by-nc/4.0/ or send a letter to Creative
# Commons, PO Box 1866, Mountain View, CA 94042, USA.


import unittest

import numpy as np
import pandas as pd

from ..statistics import descriptive


__copyright__ = "Copyright 2016, Beaulieu-Saucier Pharmacogenomics Centre"
__license__ = "Attribution-NonCommercial 4.0 International (CC BY-NC 4.0)"


class TestDescriptiveStats(unittest.TestCase):
    """Tests the descriptive statistics module."""
    def test_get_maf_without_null(self):
        """Tests when there are no missing genotypes."""
        genotypes = pd.Series([0, 0, 1, 0, 1, 0, 2, 0, 1, 0])
        maf, minor, major, flip = descriptive.get_maf(
            genotypes, "A", "B",
        )
        self.assertAlmostEqual(5 / 20, maf, places=10)
        self.assertEqual("A", minor)
        self.assertEqual("B", major)
        self.assertFalse(flip)

    def test_get_maf_flip(self):
        """Tests when there are no missing genotypes."""
        genotypes = pd.Series([2, 2, 1, 2, 1, 2, 0, 2, 1, 2])
        maf, minor, major, flip = descriptive.get_maf(
            genotypes, "A", "B",
        )
        self.assertAlmostEqual(5 / 20, maf, places=10)
        self.assertEqual("B", minor)
        self.assertEqual("A", major)
        self.assertTrue(flip)

    def test_get_maf_with_null(self):
        """Tests when there are missing genotypes."""
        genotypes = pd.Series([0, 0, 1, 0, np.nan, 0, np.nan, 0, 1, 0])
        maf, minor, major, flip = descriptive.get_maf(
            genotypes, "A", "B",
        )
        self.assertAlmostEqual(2 / 16, maf, places=10)
        self.assertEqual("A", minor)
        self.assertEqual("B", major)
        self.assertFalse(flip)

    def test_get_maf_no_valid_genotypes(self):
        """Tests when all genotypes are missing."""
        genotypes = pd.Series([np.nan, np.nan, np.nan, np.nan, np.nan, np.nan,
                               np.nan, np.nan, np.nan, np.nan])
        maf, minor, major, flip = descriptive.get_maf(
            genotypes, "A", "B",
        )
        self.assertTrue(np.isnan(maf))
        self.assertEqual("A", minor)
        self.assertEqual("B", major)
        self.assertFalse(flip)



"""
Utilities to build statistical models.
"""

# This file is part of genetest.
#
# This work is licensed under the Creative Commons Attribution-NonCommercial
# 4.0 International License. To view a copy of this license, visit
# http://creativecommons.org/licenses/by-nc/4.0/ or send a letter to Creative
# Commons, PO Box 1866, Mountain View, CA 94042, USA.


from ..statistics.descriptive import get_maf


__all__ = ["MAFFilter", "NameFilter"]


class VariantPredicate(object):
    def __init__(self):
        """Initialize a callable that will serve as a variant filtering
        function.

        The predicate should return True if the variant is to be analyzed and
        False otherwise.

        Variant predicates can also raise StopIteration to stop pushing SNPs.

        """
        pass


class MAFFilter(VariantPredicate):
    def __init__(self, maf):
        """Filters variants with a MAF under the specified threshold.

        Args:
            maf (float): The MAF threshold.

        """
        self.maf = maf

    def __call__(self, snp):
        """Executes the filter.

        Args:
            snp (): The variant object.

        Returns:
            bool: ``True`` if MAF >= threshold, ``False`` otherwise.

        """
        maf = get_maf(snp.genotypes.values, None, None)[0]
        return maf >= self.maf


class NameFilter(VariantPredicate):
    def __init__(self, extract):
        """Filters variant using a set of variant name.

        Args:
            extract (set): A set of variant names to extract.

        """
        self.extract = extract

    def __call__(self, snp):
        """Execute the filter.

        Args:
            snp (geneparse.Genotypes): The variant object.

        Returns:
            bool: ``True`` if the variant is in the list to extract, ``False``
            otherwise.

        """
        return snp.variant.name in self.extract



"""
"""

# This file is part of genetest.
#
# This work is licensed under the Creative Commons Attribution-NonCommercial
# 4.0 International License. To view a copy of this license, visit
# http://creativecommons.org/licenses/by-nc/4.0/ or send a letter to Creative
# Commons, PO Box 1866, Mountain View, CA 94042, USA.


import pandas as pd

from .core import PhenotypesContainer


__copyright__ = "Copyright 2016, Beaulieu-Saucier Pharmacogenomics Centre"
__license__ = "Attribution-NonCommercial 4.0 International (CC BY-NC 4.0)"


__all__ = ["TextPhenotypes"]


class TextPhenotypes(PhenotypesContainer):
    def __init__(self, filename, sample_column="sample", field_separator="\t",
                 missing_values=None, repeated_measurements=False,
                 keep_sample_column=False):
        """Instantiate a new TextPhenotypes object.

        Args:
            filename (str): The name of the text file containing the
                            phenotypes.
            sample_column (str): The name of the column containing the sample
                                 identification number (to fit with the
                                 genotypes).
            field_separator (str): The field separator (default is tabulation).
            missing_values (str or list or dict): The missing value(s).
            repeated_measurements (bool): Are the data containing repeated
                                          measurements (e.g. for MixedLM).

        """
        # TODO: Check the different column names
        self._phenotypes = pd.read_csv(filename, sep=field_separator,
                                       na_values=missing_values)

        # If there are repeated measurements, the sample column will have
        # duplicated values. We will set the index, but we won't verify its
        # integrity. Otherwise, we will check the index's integrity.
        self._repeated = repeated_measurements
        self._phenotypes = self._phenotypes.set_index(
            sample_column,
            verify_integrity=not repeated_measurements,
            drop=not keep_sample_column,
        )

        # Renaming the index
        self._phenotypes.index.name = "sample_id"

    def close(self):
        pass

    def __repr__(self):
        """The string representation."""
        return "TextPhenotypes({:,d} samples, {:,d} variables)".format(
            self.get_nb_samples(),
            self.get_nb_variables(),
        )

    def merge(self, other):
        """Merge this instance with another."""
        self._phenotypes = self._phenotypes.join(other._phenotypes)

    def get_phenotypes(self, li=None):
        """Returns a dataframe of phenotypes.

        Returns:
            pandas.DataFrame: A dataframe containing the phenotypes (with the
            sample IDs as index).

        """
        if li is None:
            return self._phenotypes

        # Check that all requested variables are available.
        missing_variables = set(li) - set(self._phenotypes.columns)
        if missing_variables:
            raise KeyError(
                "Some of the requested phenotypes are unavailable in the "
                "PhenotypeContainer: {}.".format(missing_variables)
            )

        return self._phenotypes.loc[:, list(li)]

    def get_nb_samples(self):
        """Returns the number of samples.

        Returns:
            int: The number of samples.

        """
        if self._repeated:
            # There are duplicated samples, so we count the number of unique
            # samples
            return self._phenotypes.index.unique().shape[0]
        else:
            return self._phenotypes.shape[0]

    def get_nb_variables(self):
        """Returns the number of variables.

        Returns:
            int: The number of variables.

        """
        return self._phenotypes.shape[1]

    def is_repeated(self):
        """Check if the phenotypes contain repeated measurements.

        Returns:
            bool: ``True`` if the data contains repeated measurements,
            ``False`` otherwise.

        """
        return self._repeated

    def keep_samples(self, keep):
        """Keeps only a subset of samples.

        Args:
            keep (set): The list of samples to keep.

        """
        self._phenotypes = self._phenotypes[self._phenotypes.index.isin(keep)]
